# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License. 
Created on 2025/6/25 10:20
Email: yundi.xxii@outlook.com
Description: 
---------------------------------------------
"""

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .data import DataLoader
    from .workflow import WorkFlow
else:
    import ygo
    DataLoader = ygo.lazy_import("quda.ml.data.DataLoader")
    WorkFlow = ygo.lazy_import("quda.ml.workflow.WorkFlow")

__version__ = "0.2.16"

__all__ = [
    "WorkFlow",
    "DataLoader",
]




# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/6/16 15:42
@author: ZhangYundi
@email: yundi.xxii@outlook.com
@description: 
---------------------------------------------
"""

from .stats import (
    aggregate_returns,
    alpha,
    alpha_aligned,
    alpha_beta,
    alpha_beta_aligned,
    annual_return,
    annual_volatility,
    beta,
    beta_aligned,
    cagr,
    beta_fragility_heuristic,
    beta_fragility_heuristic_aligned,
    gpd_risk_estimates,
    gpd_risk_estimates_aligned,
    calmar_ratio,
    capture,
    conditional_value_at_risk,
    cum_returns,
    cum_returns_final,
    down_alpha_beta,
    down_capture,
    downside_risk,
    excess_sharpe,
    max_drawdown,
    omega_ratio,
    roll_alpha,
    roll_alpha_aligned,
    roll_alpha_beta,
    roll_alpha_beta,
    roll_alpha_beta_aligned,
    roll_annual_volatility,
    roll_beta,
    roll_beta_aligned,
    roll_down_capture,
    roll_max_drawdown,
    roll_sharpe_ratio,
    roll_sortino_ratio,
    roll_up_capture,
    roll_up_down_capture,
    sharpe_ratio,
    simple_returns,
    sortino_ratio,
    stability_of_timeseries,
    tail_ratio,
    up_alpha_beta,
    up_capture,
    up_down_capture,
    value_at_risk,
)

from .periods import (
    DAILY,
    WEEKLY,
    MONTHLY,
    QUARTERLY,
    YEARLY
)



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License. 
Created on 2025/7/4 10:38
Email: yundi.xxii@outlook.com
Description: 
---------------------------------------------
"""

from .core.store.database import (
    NAME,
    DB_PATH,
    CONFIG_PATH,
    get_settings,
    sql,
    put,
    has,
    tb_path,
    read_ck,
    read_mysql,
)
from .factor import Factor
from .core.qdf import from_polars, to_lazy

__version__ = "1.1.15a4"

def update():

    """
    通过全局配置文件中的`UPDATES`配置项来更新数据，需要遵循统一编写标准。每个数据源需要定义 `CONFIG` 字典和 `fetch_fn` 方法。

    添加新的更新任务步骤：

    - **Step 1**: 实现具体的数据更新逻辑

    创建一个新的 Python 文件（如：`my_project/data/update/update_news.py`），

    并在其中定义如下内容：

    >>> import quda
    >>>
    >>> CONFIG = {'news/tb1': 'SELECT * FROM source_table'}  # SQL 查询语句或其它来源标识
    >>>
    >>> def fetch_fn(tb_name, db_conf):
    ...     query = CONFIG.get(tb_name)
    ...     return quda.read_mysql(query, db_conf=db_conf)

    - **Step 2**: 在配置文件中添加配置

    在 {quda.DB_PATH}/conf/settings.toml 添加配置

    [UPDATES.news_data]

    mod = "my_project.data.update.update_news"

    update_time = "16:30"

    db_conf = "DATABASES.mysql"

    mode = "auto"

    beg_date = "2020-01-01"

    """

    from .core import updater
    import ygo
    import ylog

    update_settings = get_settings().get("UPDATES")
    if not update_settings:
        ylog.warning(f"Missing provider update configuration.")
        return
    for task, task_conf in update_settings.items():
        ylog.info(f"[{task} config]: {task_conf}")
        mod = ygo.module_from_str(task_conf["mod"])
        for tb_name in mod.CONFIG.keys():
            ygo.delay(updater.submit)(tb_name=tb_name,
                                      fetch_fn=ygo.delay(mod.fetch_fn)(db_conf=task_conf.get("db_conf")),
                                      **task_conf)()
    updater.do(debug_mode=True)


__all__ = [
    "NAME",
    "DB_PATH",
    "CONFIG_PATH",
    "get_settings",
    "sql",
    "put",
    "has",
    "tb_path",
    "read_ck",
    "read_mysql",
    "from_polars",
    "Factor",
]



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License. 
Created on 2025/6/26 15:52
Email: yundi.xxii@outlook.com
Description: 应用层
---------------------------------------------
"""

def init_config():

    """初始化配置文件"""

    from pathlib import Path
    import shutil
    from importlib.resources import files
    import ylog

    source_conf_dir = files("quda.ml").joinpath("conf")
    # 当前工作目录下的 conf/
    target_conf_dir = Path.cwd() / "conf"
    if target_conf_dir.exists():
        ylog.warning(f"[QUDA-ML] - {target_conf_dir} already exists.")
    try:
        shutil.copytree(source_conf_dir, target_conf_dir)
        ylog.info(f"[QUDA-ML] - {target_conf_dir} created.")
    except Exception as e:
        ylog.error(f"[QUDA-ML] - Failed to create {target_conf_dir}\n{e}")



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/6/6 15:34
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""

from .core import Factor
from .consts import FIELD_DATE, FIELD_TIME, FIELD_ASSET, INDEX, DATE_FORMAT, TIME_FORMAT


# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/6/23 09:49
@author: ZhangYundi
@email: yundi.xxii@outlook.com
@description: 
---------------------------------------------
"""

from .factory import (
    Function,
    Cast,
    Expression,
    Imputer,
    Replace,
    Target,
    FilterNotNull,
    TargetFromDifferentTag,
    Reindex,
)

from .partial import StandardScaler

__all__ = [
    "Function",
    "Cast",
    "Expression",
    "Imputer",
    "Replace",
    "Target",
    "FilterNotNull",
    "StandardScaler",
    "TargetFromDifferentTag",
    "Reindex",
]


# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License.
Created on 2025/6/25 15:05
Email: yundi.xxii@outlook.com
Description: 主流程控制
---------------------------------------------
"""

import ygo
from omegaconf import DictConfig, OmegaConf
from sklearn.base import BaseEstimator
from sklearn.pipeline import Pipeline

from ..data import DataLoader
from ..pipeline import IncrementalPipeline
from ..trainer import BaseIncrementalTrainer
from ..utils import is_notebook, hydra_wrapper
from .evaluator import train, predict


class WorkFlow(BaseEstimator):
    """
    构建 训练工作流: 从数据加载->预处理->增量训练。支持 sklearn式更改配置、从配置文件中加载
    """

    def __init__(self,
                 data: list,
                 trainer: BaseIncrementalTrainer,
                 preprocessor: list | None = None,
                 trainer_params: dict | None = None):
        """
        构造工作流

        Parameters
        ----------
        data: list
            数据加载步骤
        trainer: BaseIncrementalTrainer
            训练器：quda.ml.trainer.BaseIncrementalTrainer 的实现实例
        preprocessor: list
            预处理步骤，默认None: 不做预处理
        trainer_params:
            训练器参数
        """
        # 初始化
        self.data = DataLoader(steps=data)
        self.trainer = trainer
        if isinstance(self.trainer, type):
            self.trainer = self.trainer()
        assert isinstance(self.trainer,
                          BaseIncrementalTrainer), "trainer must be a subclass of `BaseIncrementalTrainer`"
        self.trainer_params = trainer_params
        self.preprocessor = IncrementalPipeline(preprocessor) if preprocessor else None
        self.pipe = Pipeline([
            ("data", self.data.pipe),
            ("trainer", self.trainer)
        ])
        if self.preprocessor:
            self.pipe.steps.insert(1, ("preprocessor", self.preprocessor.pipe))

        self._html_repr = self.pipe._html_repr

    @classmethod
    def initialize(cls, cfg: DictConfig) -> 'WorkFlow':
        """
        从 解析出来的 初始化 workflow
        """

        from hydra.utils import instantiate

        # 构建 data pipeline
        data_steps = []
        data_steps = instantiate(cfg.data)
        data_steps = [(name, step) for d in data_steps for name, step in d.items()]

        # 构建 preprocessor
        preprocessor_steps = None
        if cfg.get("preprocessor"):
            preprocessor_steps = []
            preprocessor_steps = instantiate(cfg.preprocessor)
            preprocessor_steps = [(name, step) for d in preprocessor_steps for name, step in d.items()]

        # 构建 trainer
        trainer = instantiate(cfg.trainer.trainer)

        # trainer_params
        trainer_params = instantiate(cfg.trainer.trainer_params)
        trainer_params = OmegaConf.to_container(trainer_params, resolve=True)

        wf = cls(data=data_steps,
                 trainer=trainer,
                 preprocessor=preprocessor_steps,
                 trainer_params=trainer_params)
        return wf

    @classmethod
    def from_conf(cls, config_path: str = "conf", config_name: str = "config") -> 'WorkFlow':
        """
        配置驱动:从配置文件中构建工作流


        Examples
        --------
        >>> import quda.ml
        >>> workflow = quda.ml.WorkFlow.from_conf(config_path='conf', config_name="config")
        >>> workflow.set_params(...) # 更改配置
        >>> workflow.fit(...) # 开始训练

        Returns
        -------
        """
        if is_notebook():
            print(f"""Run in notebook, please use hydra to run.
::code::
from hydra import initialize, compose
with initialize(version_base=None, config_path='{config_path}'):
    cfg = compose(config_name='{config_name}')
wf = quda.ml.WorkFlow.initialize(cfg)
""")
            return

        hydra_main = hydra_wrapper(config_path, config_name)
        if hydra_main is not None:
            return hydra_main(cls.initialize)()

    @classmethod
    def train(cls, cfg: DictConfig):
        """
        从解析后的config运行配置
        Returns
        -------
        """
        wf = cls.initialize(cfg)
        sig_params = ygo.fn_signature_params(train)
        params = {k: v for k, v in cfg.items() if k in sig_params}
        if params:
            from hydra.utils import instantiate
            params = instantiate(params)
        return train(workflow=wf, **params)

    @classmethod
    def train_conf(cls, config_path: str = "conf", config_name: str = "config", ):
        """
        从配置文件中创建 WorkFlow 对象，并且运行 train

        Returns
        -------

        """

        if is_notebook():
            print(f"""Run in notebook, please use hydra to run.
::code::
from hydra import initialize, compose
with initialize(version_base=None, config_path='{config_path}'):
    cfg = compose(config_name='{config_name}')
quda.ml.WorkFlow.train(cfg)
""")
            return
        hydra_main = hydra_wrapper(config_path, config_name)
        if hydra_main is not None:
            return hydra_main(cls.train)()

    @classmethod
    def predict(cls, cfg: DictConfig):
        """
        从解析后的config运行配置

        Returns
        -------
        polars.DataFrame
        """
        data_loader = DataLoader.initialize(cfg)
        sig_params = ygo.fn_signature_params(predict)
        params = {k: v for k, v in cfg.items() if k in sig_params}
        if params:
            from hydra.utils import instantiate
            params = instantiate(params)
        return predict(data_loader=data_loader, **params)

    @classmethod
    def predict_conf(cls, config_path: str = "conf", config_name: str = "config",):
        """
        从配置文件中运行预测

        Parameters
        ----------
        config_path: str
        config_name: str

        Returns
        -------
        polars.DataFrame
        """
        if is_notebook():
            print(f"""Run in notebook, please use hydra to run.
::code::
from hydra import initialize, compose
with initialize(version_base=None, config_path='{config_path}'):
    cfg = compose(config_name='{config_name}')
quda.ml.WorkFlow.predict(cfg)
""")
            return
        hydra_main = hydra_wrapper(config_path, config_name)
        if hydra_main is not None:
            return hydra_main(cls.predict)()



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License. 
Created on 2025/7/2 14:46
Email: yundi.xxii@outlook.com
Description: 
---------------------------------------------
"""

from typing import Sequence, TYPE_CHECKING

import mlflow
import ylog

from ..utils import extract_feature_names, get_timedelta_names
from sklearn.model_selection import train_test_split

if TYPE_CHECKING:
    from .workflow import WorkFlow
    from ..data import DataLoader


def train(workflow: 'WorkFlow',
          iterables: Sequence,
          batch_size: int = -1,
          features: list[str] | None = None,
          valid_size: float = 0.2,
          shuffle: bool = False,
          random_state: int = 42, ):
    """
    通用增量训练流程，支持分批加载数据,预处理，模型训练与保存

    Parameters
    ----------
    workflow: 'WorkFlow'
    iterables: Sequence
        可迭代对象
    batch_size: int
        每次训练的批次，默认 -1: 全量训练
    features: list[str]|None
        特征名, 默认None。会排除date,time,asset,price,以timedelta命名的收益列
    valid_size: float
        验证集比例，默认0.2
    shuffle: bool
        是否打乱 iterables, 默认 False
    random_state: int
        随机数种子，默认 42

    Returns
    -------

    """
    batch_size = batch_size if batch_size > 0 else len(iterables)
    iterables = sorted(iterables)

    mlflow.autolog()
    mlflow.sklearn.autolog(disable=True)

    with mlflow.start_run() as top_run:
        top_run_id = top_run.info.run_id
        ylog.info(f"Started mlflow.run: {top_run_id}")
        params = {
            "data.beg": iterables[0],
            "data.end": iterables[-1],
            "batch_size": batch_size,
            "shuffle": shuffle,
            "random_state": random_state,
        }
        mlflow.log_params(params, run_id=top_run_id)
        train_batch_list = [iterables[i:i + batch_size] for i in range(0, len(iterables), batch_size)]
        for i, train_batch in enumerate(train_batch_list):
            beg, end = train_batch[0], train_batch[-1]
            evals_result = {}
            with mlflow.start_run(nested=True, ) as partial_run:
                partial_run_id = partial_run.info.run_id
                ylog.info(f"Started mlflow.run: {partial_run_id}")
                ylog.info(f"Preparing on batch_{i + 1}({i + 1}/{len(train_batch_list)}) - {beg} >>> {end}")
                params["data.beg"] = beg
                params["data.end"] = end
                mlflow.log_params(params, run_id=partial_run_id)
                # 加载训练数据
                train_list_, valid_list_ = train_test_split(train_batch,
                                                            test_size=valid_size,
                                                            shuffle=shuffle,
                                                            random_state=random_state)
                train_dl = workflow.data.fetch(train_list_,
                                               batch_size=-1,
                                               eager=True)
                train_data = next(train_dl)
                if workflow.preprocessor:
                    train_data = workflow.preprocessor.partial_fit_transform(train_data)
                if valid_list_:
                    valid_dl = workflow.data.fetch(valid_list_,
                                                   batch_size=-1,
                                                   eager=True)
                    valid_data = next(valid_dl)
                    if workflow.preprocessor:
                        valid_data = workflow.preprocessor.fit_transform(valid_data)
                else:
                    valid_data = None

                if not features:
                    features = extract_feature_names(train_data.columns)
                else:
                    features = features
                if workflow.trainer.input_example is None:
                    workflow.trainer.input_example = train_data[features].head().to_pandas()
                target = get_timedelta_names(train_data.columns)
                if len(target) >= 1:
                    ylog.info(f"Target column found: {target}, use {target[0]}")
                else:
                    raise ValueError("Miss target columns in data")

                target = target[0]

                ylog.info("Building dataset.train...")
                train_data = workflow.trainer.create_dataset(data=train_data,
                                                             features=features,
                                                             target=target, )

                if valid_data is not None:
                    ylog.info("Building dataset.valid...")
                    valid_data = workflow.trainer.create_dataset(data=valid_data,
                                                                 features=features,
                                                                 target=target, )

                ylog.info(f"Training on batch_{i + 1}({i + 1}/{len(train_batch_list)}) - {beg} >>> {end}")

                # 训练模型
                workflow.trainer.fit(train_data=train_data,
                                     valid_data=valid_data,
                                     init_model=workflow.trainer.model,
                                     evals_result=evals_result,
                                     **workflow.trainer_params)

                # 保存模型
                # model_name = f"checkpoint_{i + 1}"
                # model_info = self.trainer.save_model(model, model_name, )
                # self.trainer.model = mlflow.pyfunc.load_model(model_info.model_uri)
                workflow.trainer.model = mlflow.pyfunc.load_model(f"runs:/{partial_run_id}/model")

                model = None


def predict(data_loader: 'DataLoader',
            model_uri: str,
            iterables: Sequence,
            batch_size: int = -1,
            features: list[str] | None = None,
            index: tuple[str] = ("date", "time", "asset"), ):
    """
    模型预测

    Parameters
    ----------
    data_loader: 'DataLoader',

    model_uri: str

    iterables: Sequence
        待评估的序列
    batch_size: int
        每次读取数据的批次，默认 -1: 全量读取
    features: list[str]|None
        特征名, 默认None。会排除date,time,asset,price,以timedelta命名的收益列
    index: tuple[str]
    Returns
    -------

    """
    import polars as pl

    batch_size = batch_size if batch_size > 0 else len(iterables)
    iterables = sorted(iterables)
    test_batch_list = [iterables[i:i + batch_size] for i in range(0, len(iterables), batch_size)]
    mod = mlflow.pyfunc.load_model(model_uri)
    preds = list()
    for i, test_batch in enumerate(test_batch_list):
        beg, end = test_batch[0], test_batch[-1]
        ylog.info(f"Preparing on batch_{i + 1}({i + 1}/{len(test_batch_list)}) - {beg} >>> {end}")
        test_dl = data_loader.fetch(test_batch, batch_size=-1,)
        test_data = next(test_dl)
        if not features:
            features = extract_feature_names(test_data.columns)
        else:
            features = features

        timedelta_names = get_timedelta_names(test_data.columns)

        ylog.info(f"Predicting on batch_{i + 1}({i + 1}/{len(test_batch_list)}) - {beg} >>> {end}")
        pred = mod.predict(test_data[features].to_numpy())
        pred = pl.Series(name="pred", values=pred)
        pred_batch = test_data.select(index).with_columns(pred)
        for timedelta_name in timedelta_names:
            pred_batch = pred_batch.with_columns(test_data[timedelta_name])
        preds.append(pred_batch)
    if preds:
        pred_df = pl.concat(preds)
    else:
        pred_df = None
    return pred_df



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/6/23 09:49
@author: ZhangYundi
@email: yundi.xxii@outlook.com
@description: 基础转换器仓库
---------------------------------------------
"""

import inspect
import math
from collections.abc import Sequence, Iterable
from typing import Literal

import numpy as np
import polars as pl
import ygo
from sklearn.base import BaseEstimator, TransformerMixin

SPECIAL_VALUES = {
    "null": None,
    "none": None,
    "nan": None,
    "inf": np.inf,
    "-inf": -np.inf,
    "int": int,
}


class Function(BaseEstimator, TransformerMixin):
    """
    将任意函数封装为支持 set_params 的 BaseEstimator

    Parameters
    ----------
    fn: callable
        方法对象
    kwargs: dict
        fn的参数

    Examples
    --------
    >>> import quda
    >>> func = Function(lambda date, ds_path: quda.sql(f"select * from {ds_path} where date='{date}'"))
    >>> func.set_params(ds_path="mc/stock_kline_day")
    """

    def __init__(self, fn: callable, **kwargs):
        self.fn = ygo.delay(fn)(**kwargs)
        self.params = {k: inspect.Parameter.empty for k in ygo.fn_signature_params(self.fn)}
        self.params.update(self.fn.stored_kwargs)

    def set_params(self, **params):
        self.fn = ygo.delay(self.fn)(**params)
        self.params.update(self.fn.stored_kwargs)
        return self

    def get_params(self, deep=True):
        return self.params

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        if self.fn is None:
            raise ValueError("Function not provided.")
        params = {k: v for k, v in self.fn.stored_kwargs.items() if v is not inspect.Parameter.empty}
        return self.fn(X, **params)


class FilterNotNull(BaseEstimator, TransformerMixin):
    """
    Drop all rows that contain one or more null values

    Parameters
    ----------
    subset: Sequence[str] | str | None
        剔除空值的字段

    """

    def __init__(self, subset: Sequence[str] | str | None = None):
        self.subset = subset

    def fit(self, X=None, y=None):
        return self

    def transform(self, X: pl.LazyFrame):
        if self.subset:
            return X.drop_nulls(pl.sql_expr(self.subset))
        return X


class Cast(BaseEstimator, TransformerMixin):
    """
    类型转换

    Parameters
    ----------
    old: str | polars.DataType | polars.Expr
        原本类型或者需要转换的列
    new: polars.DataType
        转换的目标类型

    Examples
    --------
    >>> Cast("asset", pl.UInt16)

    """

    def __init__(self, old: pl.Expr | str | pl.DataType = None, new=None):
        self.old = old
        self.new = new
        self.old_expr: pl.Expr = None

    def fit(self, X, y=None):
        if isinstance(self.old, str):
            if self.old in SPECIAL_VALUES:
                self.old = SPECIAL_VALUES[self.old]
        self.old_expr = self.old if isinstance(self.old, pl.Expr) else pl.col(self.old)
        return self

    def transform(self, X):
        return X.with_columns(self.old_expr.cast(self.new))


class Expression(BaseEstimator, TransformerMixin):
    """
    expression engine

    Parameters
    ----------
    exprs: list[str]

    """

    def __init__(self, exprs: list[str] | None = None, index: tuple[str] = ("date", "time", "asset"),
                 align: bool = False):
        self.index = index
        self.exprs = exprs
        self.align = align

    def fit(self, X, y=None):
        return self

    def transform(self, X):
        import quda
        return quda.to_lazy(X, self.index, self.align).sql(*self.exprs)


class Imputer(BaseEstimator, TransformerMixin):
    """
    插值器

    Parameters
    ----------
    strategy
        插值策略
    over_spec
        用于滚动插值
    columns
        需要处理的列
    """

    def __init__(self,
                 strategy: Literal[
                               "forward", "backward", "mean", "zero", "max", "min", "one"] | None = "forward",
                 over_spec: dict | None = None,
                 columns: Sequence[str] | None = None):
        self.strategy = strategy
        self.over_spec = over_spec
        self.columns = [columns, ] if isinstance(columns, str) else columns
        self._exprs: list[pl.Expr] = list()

    def fit(self, X=None, y=None):
        return self

    def transform(self, X: pl.LazyFrame):
        cols = X.collect_schema().names() if not self.columns else self.columns
        if not self.over_spec:
            self._exprs = [pl.col(col).fill_null(strategy=self.strategy) for col in cols]
        else:
            self._exprs = [pl.col(col).fill_null(strategy=self.strategy).over(**self.over_spec) for col in self.columns]
        return X.with_columns(self._exprs)


class Replace(BaseEstimator, TransformerMixin):
    """
    替换器

    Parameters
    ----------
    columns: list
        需要替换的目标列
    old
        旧值, 可以是list, 如[np.inf, -np.inf]
    new
        新值
    """

    def __init__(self, columns: list | None = None, old=None, new=None):

        self.columns = columns
        self.old = old
        self.new = new

    def fit(self, X, y=None):
        if isinstance(self.old, str):
            if self.old in SPECIAL_VALUES:
                self.old = SPECIAL_VALUES[self.old]
        elif isinstance(self.old, Iterable):
            self.old = [SPECIAL_VALUES[old] if old in SPECIAL_VALUES else old for old in self.old]
        if self.columns is None:
            self.columns = list()
        return self

    def transform(self, X):
        return X.with_columns(pl.col(c).replace(self.old, self.new) for c in self.columns)


class Target(BaseEstimator, TransformerMixin):
    """
    添加目标收益列

    Parameters
    ----------
    price_tag
        价格列，支持sql语法，如 `if(ask_p1 > 0, bid_p1 > 0, (ask_p1+bid_p1)/2, null) as price`
    frequency
        数据集频率, 默认`3s`
    target
        目标收益，默认`5min`
    partition_by
        分区
    order_by
        排序
    """

    def __init__(self,
                 price_tag: str = "price",
                 frequency: str = "3s",
                 target: str = "5min",
                 gap: str = "3s",
                 partition_by: Sequence[str] | str = "asset",
                 order_by: Sequence[str] | str | None = "time",
                 alias: str = None):
        self.price_tag = price_tag
        self.frequency = frequency
        self.target = target
        self.gap = gap
        self.partition_by = partition_by
        self.order_by = order_by
        self.alias = alias
        self._exprs = list()

    def fit(self, X=None, y=None):
        from pandas import Timedelta
        self.alias = self.alias if self.alias else self.target
        over_spec = {"partition_by": self.partition_by}
        if self.order_by:
            over_spec["order_by"] = self.order_by
        freq_secs = int(Timedelta(self.frequency).seconds)
        expr_target = (
            (
                    pl.sql_expr(self.price_tag)
                    .shift(-math.ceil(Timedelta(self.target).seconds / freq_secs))
                    .over(**over_spec) / pl.sql_expr(self.price_tag) - 1
            ).alias(self.alias)
        )
        expr_inf = pl.col(self.alias).cast(pl.Float32).replace([float("inf"), float("-inf")], None)

        self._exprs = [expr_target, expr_inf]
        gap_secs = int(Timedelta(self.gap).seconds)
        if gap_secs > 0:
            expr_gap = (
                pl.col(self.alias)
                .shift(-math.ceil(gap_secs / freq_secs))
                .over(**over_spec)
            )
            self._exprs.insert(1, expr_gap)
        return self

    def transform(self, X: pl.LazyFrame):
        for expr in self._exprs:
            X = X.with_columns(expr)
        return X


class TargetFromDifferentTag(BaseEstimator, TransformerMixin):
    """
    添加目标收益列: 使用不同的价格列

    Parameters
    ----------
    enter_price_tag: str
        进场价格列，支持sql语法，如 `if(ask_p1 > 0, bid_p1 > 0, (ask_p1+bid_p1)/2, null) as price`
    exit_price_tag: str
        出场价格列
    frequency
        数据集频率, 默认`3s`
    target
        目标收益，默认`5min`
    partition_by
        分区
    order_by
        排序
    alias: str
        重新命名
    """

    def __init__(self,
                 enter_price_tag: str,
                 exit_price_tag: str,
                 frequency: str = "3s",
                 target: str = "5min",
                 gap: str = "3s",
                 partition_by: Sequence[str] | str = "asset",
                 order_by: Sequence[str] | str | None = "time",
                 alias: str = None):
        self.enter_price_tag = enter_price_tag
        self.exit_price_tag = exit_price_tag
        self.frequency = frequency
        self.target = target
        self.gap = gap
        self.partition_by = partition_by
        self.order_by = order_by
        self.alias = alias
        self._exprs = list()

    def fit(self, X=None, y=None):
        from pandas import Timedelta
        self.alias = self.alias if self.alias else self.target
        over_spec = {"partition_by": self.partition_by}
        if self.order_by:
            over_spec["order_by"] = self.order_by
        freq_secs = int(Timedelta(self.frequency).seconds)
        expr_target = (
            (
                    pl.sql_expr(self.exit_price_tag)
                    .shift(-math.ceil(Timedelta(self.target).seconds / freq_secs))
                    .over(**over_spec) / pl.sql_expr(self.enter_price_tag) - 1
            ).alias(self.alias)
        )
        expr_inf = pl.col(self.alias).cast(pl.Float32).replace([float("inf"), float("-inf")], None)

        self._exprs = [expr_target, expr_inf]
        gap_secs = int(Timedelta(self.gap).seconds)
        if gap_secs > 0:
            expr_gap = (
                pl.col(self.alias)
                .shift(-math.ceil(gap_secs / freq_secs))
                .over(**over_spec)
            )
            self._exprs.insert(1, expr_gap)
        return self

    def transform(self, X: pl.LazyFrame):
        for expr in self._exprs:
            X = X.with_columns(expr)
        return X

class Reindex(BaseEstimator, TransformerMixin):

    def __init__(self, new_index: pl.LazyFrame | pl.DataFrame):
        self.new_index = new_index

    def fit(self, X, y=None):
        self.new_index = self.new_index.lazy()
        return self

    def transform(self, X: pl.LazyFrame):
        return self.new_index.join(X, on=self.new_index.collect_schema().names(), how="left")


# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/5/29 18:10
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""

def sql_StockKlineDay(date, tb_name):
    return f"""
    select date,
           replaceRegexpAll(order_book_id, '[^0-9]', '')                          as asset,
           total_turnover                                                         as amount,
           volume,
           prev_close,
           open,
           high,
           low,
           close,
           limit_up,
           limit_down,
           if(num_trades < 0, 0, if(num_trades > toInt64(volume), 0, num_trades)) as num_trades
    from {tb_name}
        prewhere date = '{date}'
    order by asset
    """


def sql_StockKlineMinute(date, tb_name):
    return f"""
    select EventDate                                                              as date,
           replaceRegexpAll(order_book_id, '[^0-9]', '')                          as asset,
           formatDateTime(datetime, '%T')                                         as time,
           total_turnover                                                         as amount,
           volume,
           open,
           high,
           low,
           close,
           if(num_trades < 0, 0, if(num_trades > toInt64(volume), 0, num_trades)) as num_trades
    from {tb_name}
        prewhere EventDate = '{date}'
    order by asset
    """

def sql_StockTick(date, tb_name):
    return f"""
    SELECT EventDate as date,
               formatDateTime(datetime, '%T') as time,
               replaceRegexpAll(order_book_id, '[^0-9]', '') AS asset,
               total_turnover AS amount,
               volume,
               high,
               low,
               close,
               a1 AS ask_p1,
               a2 AS ask_p2,
               a3 AS ask_p3,
               a4 AS ask_p4,
               a5 AS ask_p5,
               a1_v AS ask_v1,
               a2_v AS ask_v2,
               a3_v AS ask_v3,
               a4_v AS ask_v4,
               a5_v AS ask_v5,
               b1 AS bid_p1,
               b2 AS bid_p2,
               b3 AS bid_p3,
               b4 AS bid_p4,
               b5 AS bid_p5,
               b1_v AS bid_v1,
               b2_v AS bid_v2,
               b3_v AS bid_v3,
               b4_v AS bid_v4,
               b5_v AS bid_v5
        FROM {tb_name}
        WHERE EventDate = '{date}'
        AND time >= '09:14:00'
        ORDER BY
            order_book_id ASC,
            datetime ASC
    """


def sql_IndexKlineMinute(date, tb_name):
    return f"""
    select EventDate                                     as date,
           replaceRegexpAll(order_book_id, '[^0-9]', '') as asset,
           formatDateTime(datetime, '%T')                as time,
           total_turnover                                as amount,
           volume,
           open,
           high,
           low,
           close
    from {tb_name}
        prewhere EventDate = '{date}'
    order by asset
    """


def sql_IndexKlineDay(date, tb_name):
    return f"""
    select date,
           replaceRegexpAll(order_book_id, '[^0-9]', '') as asset,
           total_turnover                                as amount,
           volume,
           prev_close,
           open,
           high,
           low,
           close
    from {tb_name}
        prewhere date = '{date}'
    order by asset
    """



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/6/5 20:34
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""

import ast
import hashlib
import re
from inspect import getsource

import ygo


def remove_end_time_default(fn):
    # 获取源码并解析为 AST
    source = getsource(fn)
    module = ast.parse(source)

    # 找到函数定义节点
    func_node = next(node for node in module.body if isinstance(node, ast.FunctionDef))

    # 移除 end_time 参数
    new_args = []
    defaults = []

    arg_names = [arg.arg for arg in func_node.args.args]
    default_values = func_node.args.defaults

    # 记录有多少个参数有默认值
    num_defaults = len(default_values)
    offset = len(arg_names) - num_defaults

    for i, arg in enumerate(func_node.args.args):
        if arg.arg == 'end_time':
            continue  # 跳过 end_time 参数

        new_args.append(arg)
        # 如果该参数有默认值
        if i >= offset:
            defaults.append(default_values[i - offset])

    # 更新函数参数
    func_node.args.args = new_args
    func_node.args.defaults = defaults

    # 可选：如果 kwonlyargs 中还有 end_time，也移除
    new_kwonlyargs = []
    new_kwonlydefaults = []
    for arg, default in zip(func_node.args.kwonlyargs, func_node.args.kw_defaults):
        if arg.arg == 'end_time':
            continue
        new_kwonlyargs.append(arg)
        new_kwonlydefaults.append(default)

    func_node.args.kwonlyargs = new_kwonlyargs
    func_node.args.kw_defaults = new_kwonlydefaults

    # 重新生成源码
    new_source = ast.unparse(func_node)
    return new_source


def fn_code(fn: callable) -> str:
    """
    返回清洗后的函数定义字符串：
    - 移除 end_time 参数

    Parameters
    ----------
    fn : callable
        需要获取定义代码的callable对象

    Returns
    -------
    str
        清洗后的函数定义字符串
    """
    # 使用 AST 解析并修改函数定义
    try:
        return remove_end_time_default(fn)
    except Exception as e:
        # 如果 AST 解析失败，退回到原始方式
        return getsource(fn)


def clean_source(source):
    # 移除注释
    source = re.sub(r'#.*$', '', source, flags=re.MULTILINE)
    # 移除空行
    source = re.sub(r'\n\s*\n', '\n', source)
    return source.strip()


def fn_code_hash(fn):
    source = fn_code(fn)
    cleaned = clean_source(source)
    return hashlib.md5(cleaned.encode()).hexdigest()


def fn_params_hash(fn, ignore_params=None):
    if ignore_params is None:
        ignore_params = []
    params = ygo.fn_params(fn)
    filtered = [f'{k}={v}' for k, v in params if k not in ignore_params]
    filtered_str = f"({','.join(filtered)})"
    return hashlib.md5(filtered_str.encode()).hexdigest()


def version_hash(fn, ):
    code_hash = fn_code_hash(fn)
    param_hash = fn_params_hash(fn, ignore_params=["end_time"])
    combined = f"{code_hash}|{param_hash}"
    return hashlib.md5(combined.encode()).hexdigest()



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2024/12/9 下午3:06
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""

from .options import Options
from .plotting import (
    table,
    distplot,
    bar,
    violin,
    lines,
    nv_plot,
)

__all__ = ["table",
           "distplot",
           "bar",
           "Options",
           "violin",
           "lines",
           "nv_plot"]



"""Utilities for marking deprecated functions."""
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
from functools import wraps


def deprecated(msg=None, stacklevel=2):
    """
    Used to mark a function as deprecated.
    Parameters
    ----------
    msg : str
        The message to display in the deprecation warning.
    stacklevel : int
        How far up the stack the warning needs to go, before
        showing the relevant calling lines.
    Usage
    -----
    @deprecated(msg='function_a is deprecated! Use function_b instead.')
    def function_a(*args, **kwargs):
    """
    def deprecated_dec(fn):
        @wraps(fn)
        def wrapper(*args, **kwargs):
            warnings.warn(
                msg or "Function %s is deprecated." % fn.__name__,
                category=DeprecationWarning,
                stacklevel=stacklevel
            )
            return fn(*args, **kwargs)
        return wrapper
    return deprecated_dec



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/6/16 15:46
@author: ZhangYundi
@email: yundi.xxii@outlook.com
@description: 
---------------------------------------------
"""

APPROX_BDAYS_PER_MONTH = 21
APPROX_BDAYS_PER_YEAR = 252

MONTHS_PER_YEAR = 12
WEEKS_PER_YEAR = 52
QTRS_PER_YEAR = 4

DAILY = "daily"
WEEKLY = "weekly"
MONTHLY = "monthly"
QUARTERLY = "quarterly"
YEARLY = "yearly"

ANNUALIZATION_FACTORS = {
    DAILY: APPROX_BDAYS_PER_YEAR,
    WEEKLY: WEEKS_PER_YEAR,
    MONTHLY: MONTHS_PER_YEAR,
    QUARTERLY: QTRS_PER_YEAR,
    YEARLY: 1,
}


#
# Copyright 2018 Quantopian, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from datetime import datetime
from functools import wraps
from os import makedirs, environ
from os.path import expanduser, join, getmtime, isdir
import errno
import warnings

import numpy as np
from numpy.lib.stride_tricks import as_strided
import pandas as pd
from pandas.tseries.offsets import BDay

from .deprecate import deprecated

DATAREADER_DEPRECATION_WARNING = \
        ("Yahoo and Google Finance have suffered large API breaks with no "
         "stable replacement. As a result, any data reading functionality "
         "in empyrical has been deprecated and will be removed in a future "
         "version. See README.md for more details: "
         "\n\n"
         "\thttps://github.com/quantopian/pyfolio/blob/master/README.md")
try:
    # fast versions
    import bottleneck as bn

    def _wrap_function(f):
        @wraps(f)
        def wrapped(*args, **kwargs):
            out = kwargs.pop('out', None)
            data = f(*args, **kwargs)
            if out is None:
                out = data
            else:
                out[()] = data

            return out

        return wrapped

    nanmean = _wrap_function(bn.nanmean)
    nanstd = _wrap_function(bn.nanstd)
    nansum = _wrap_function(bn.nansum)
    nanmax = _wrap_function(bn.nanmax)
    nanmin = _wrap_function(bn.nanmin)
    nanargmax = _wrap_function(bn.nanargmax)
    nanargmin = _wrap_function(bn.nanargmin)
except ImportError:
    # slower numpy
    nanmean = np.nanmean
    nanstd = np.nanstd
    nansum = np.nansum
    nanmax = np.nanmax
    nanmin = np.nanmin
    nanargmax = np.nanargmax
    nanargmin = np.nanargmin


def roll(*args, **kwargs):
    """
    Calculates a given statistic across a rolling time period.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns (optional): float / series
        Benchmark return to compare returns against.
    function:
        the function to run for each rolling window.
    window (keyword): int
        the number of periods included in each calculation.
    (other keywords): other keywords that are required to be passed to the
        function in the 'function' argument may also be passed in.

    Returns
    -------
    np.ndarray, pd.Series
        depends on input type
        ndarray(s) ==> ndarray
        Series(s) ==> pd.Series

        A Series or ndarray of the results of the stat across the rolling
        window.

    """
    func = kwargs.pop('function')
    window = kwargs.pop('window')
    if len(args) > 2:
        raise ValueError("Cannot pass more than 2 return sets")

    if len(args) == 2:
        if not isinstance(args[0], type(args[1])):
            raise ValueError("The two returns arguments are not the same.")

    if isinstance(args[0], np.ndarray):
        return _roll_ndarray(func, window, *args, **kwargs)
    return _roll_pandas(func, window, *args, **kwargs)


def up(returns, factor_returns, **kwargs):
    """
    Calculates a given statistic filtering only positive factor return periods.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns (optional): float / series
        Benchmark return to compare returns against.
    function:
        the function to run for each rolling window.
    (other keywords): other keywords that are required to be passed to the
        function in the 'function' argument may also be passed in.

    Returns
    -------
    Same as the return of the function
    """
    func = kwargs.pop('function')
    returns = returns[factor_returns > 0]
    factor_returns = factor_returns[factor_returns > 0]
    return func(returns, factor_returns, **kwargs)


def down(returns, factor_returns, **kwargs):
    """
    Calculates a given statistic filtering only negative factor return periods.

    Parameters
    ----------
    returns : pd.Series or np.ndarray
        Daily returns of the strategy, noncumulative.
        - See full explanation in :func:`~empyrical.stats.cum_returns`.
    factor_returns (optional): float / series
        Benchmark return to compare returns against.
    function:
        the function to run for each rolling window.
    (other keywords): other keywords that are required to be passed to the
        function in the 'function' argument may also be passed in.

    Returns
    -------
    Same as the return of the 'function'
    """
    func = kwargs.pop('function')
    returns = returns[factor_returns < 0]
    factor_returns = factor_returns[factor_returns < 0]
    return func(returns, factor_returns, **kwargs)


def _roll_ndarray(func, window, *args, **kwargs):
    data = []
    for i in range(window, len(args[0]) + 1):
        rets = [s[i-window:i] for s in args]
        data.append(func(*rets, **kwargs))
    return np.array(data)


def _roll_pandas(func, window, *args, **kwargs):
    data = {}
    index_values = []
    for i in range(window, len(args[0]) + 1):
        rets = [s.iloc[i-window:i] for s in args]
        index_value = args[0].index[i - 1]
        index_values.append(index_value)
        data[index_value] = func(*rets, **kwargs)
    return pd.Series(data, index=type(args[0].index)(index_values))


@deprecated(msg=DATAREADER_DEPRECATION_WARNING)
def cache_dir(environ=environ):
    try:
        return environ['EMPYRICAL_CACHE_DIR']
    except KeyError:
        return join(

            environ.get(
                'XDG_CACHE_HOME',
                expanduser('~/.cache/'),
            ),
            'empyrical',
        )


@deprecated(msg=DATAREADER_DEPRECATION_WARNING)
def data_path(name):
    return join(cache_dir(), name)


@deprecated(msg=DATAREADER_DEPRECATION_WARNING)
def ensure_directory(path):
    """
    Ensure that a directory named "path" exists.
    """

    try:
        makedirs(path)
    except OSError as exc:
        if exc.errno != errno.EEXIST or not isdir(path):
            raise


def get_utc_timestamp(dt):
    """
    Returns the Timestamp/DatetimeIndex
    with either localized or converted to UTC.

    Parameters
    ----------
    dt : Timestamp/DatetimeIndex
        the date(s) to be converted

    Returns
    -------
    same type as input
        date(s) converted to UTC
    """

    dt = pd.to_datetime(dt)
    try:
        dt = dt.tz_localize('UTC')
    except TypeError:
        dt = dt.tz_convert('UTC')
    return dt


_1_bday = BDay()


def _1_bday_ago():
    return pd.Timestamp.now().normalize() - _1_bday



@deprecated(msg=DATAREADER_DEPRECATION_WARNING)
def get_returns_cached(filepath, update_func, latest_dt, **kwargs):
    """
    Get returns from a cached file if the cache is recent enough,
    otherwise, try to retrieve via a provided update function and
    update the cache file.
    Parameters
    ----------
    filepath : str
        Path to cached csv file
    update_func : function
        Function to call in case cache is not up-to-date.
    latest_dt : pd.Timestamp (tz=UTC)
        Latest datetime required in csv file.
    **kwargs : Keyword arguments
        Optional keyword arguments will be passed to update_func()
    Returns
    -------
    pandas.DataFrame
        DataFrame containing returns
    """

    update_cache = False

    try:
        mtime = getmtime(filepath)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise
        update_cache = True
    else:

        file_dt = pd.Timestamp(mtime, unit='s')

        if latest_dt.tzinfo:
            file_dt = file_dt.tz_localize('utc')

        if file_dt < latest_dt:
            update_cache = True
        else:
            returns = pd.read_csv(filepath, index_col=0, parse_dates=True)
            returns.index = returns.index.tz_localize("UTC")

    if update_cache:
        returns = update_func(**kwargs)
        try:
            ensure_directory(cache_dir())
        except OSError as e:
            warnings.warn(
                'could not update cache: {}. {}: {}'.format(
                    filepath, type(e).__name__, e,
                ),
                UserWarning,
            )

        try:
            returns.to_csv(filepath)
        except OSError as e:
            warnings.warn(
                'could not update cache {}. {}: {}'.format(
                    filepath, type(e).__name__, e,
                ),
                UserWarning,
            )

    return returns



def rolling_window(array, length, mutable=False):
    """
    Restride an array of shape

        (X_0, ... X_N)

    into an array of shape

        (length, X_0 - length + 1, ... X_N)

    where each slice at index i along the first axis is equivalent to

        result[i] = array[length * i:length * (i + 1)]

    Parameters
    ----------
    array : np.ndarray
        The base array.
    length : int
        Length of the synthetic first axis to generate.
    mutable : bool, optional
        Return a mutable array? The returned array shares the same memory as
        the input array. This means that writes into the returned array affect
        ``array``. The returned array also uses strides to map the same values
        to multiple indices. Writes to a single index may appear to change many
        values in the returned array.

    Returns
    -------
    out : np.ndarray

    Example
    -------
    >>> from numpy import arange
    >>> a = arange(25).reshape(5, 5)
    >>> a
    array([[ 0,  1,  2,  3,  4],
           [ 5,  6,  7,  8,  9],
           [10, 11, 12, 13, 14],
           [15, 16, 17, 18, 19],
           [20, 21, 22, 23, 24]])

    >>> rolling_window(a, 2)
    array([[[ 0,  1,  2,  3,  4],
            [ 5,  6,  7,  8,  9]],
    <BLANKLINE>
           [[ 5,  6,  7,  8,  9],
            [10, 11, 12, 13, 14]],
    <BLANKLINE>
           [[10, 11, 12, 13, 14],
            [15, 16, 17, 18, 19]],
    <BLANKLINE>
           [[15, 16, 17, 18, 19],
            [20, 21, 22, 23, 24]]])
    """
    if not length:
        raise ValueError("Can't have 0-length window")

    orig_shape = array.shape
    if not orig_shape:
        raise IndexError("Can't restride a scalar.")
    elif orig_shape[0] < length:
        raise IndexError(
            "Can't restride array of shape {shape} with"
            " a window length of {len}".format(
                shape=orig_shape,
                len=length,
            )
        )

    num_windows = (orig_shape[0] - length + 1)
    new_shape = (num_windows, length) + orig_shape[1:]

    new_strides = (array.strides[0],) + array.strides

    out = as_strided(array, new_shape, new_strides)
    out.setflags(write=mutable)
    return out



import numpy as np
import pandas as pd
import polars as pl

import plotly.graph_objs as go

import statsmodels.api as sm
import matplotlib.pyplot as plt

from scipy import stats

from typing import Sequence, Literal

import finplot
import ylog
from .graph import ScatterGraph, SubplotsGraph, BarGraph, HeatmapGraph
from ..utils import guess_plotly_rangebreaks
from ... import tears, utils


def _group_return(pred_label: pl.DataFrame = None, N: int = 5, **kwargs) -> tuple:
    """
    绘制 分组收益 图: 平均收益(直方图)/累计收益(曲线图)
    Parameters
    ----------
    pred_label
    N
    by_group
    kwargs

    Returns
    -------

    """

    # Group1 ~ Group5 only consider the dropna values
    pred_label_drop = pred_label.drop_nulls(subset=["score"])

    # Group
    grouper = ["date", "time"]

    # demean
    pred_label_demeaned = pred_label.with_columns((pl.col("label")-pl.col("label").mean().over(grouper)).alias("label"))

    long_demeaned = pred_label_demeaned.filter(pl.col("quantile")==N)

    t_df = pred_label_drop.select("date", "time", "quantile", "label").group_by("date", "quantile").mean().sort(by=["date", "quantile"]).drop("time")
    # 分组平均收益
    group_mean = t_df.group_by("quantile").mean().drop("date")
    group_mean = group_mean.with_columns((pl.col("label") * 1e4).round(2))
    # group_mean_figure = finplot.bar_plot(group_mean.to_pandas().set_index("quantile", drop=True), title="Group Mean Returns")
    group_mean = group_mean.to_pandas().set_index("quantile", drop=True)
    group_bar_figure = BarGraph(
        group_mean,
        layout=dict(
            title="Group Mean Return",
            xaxis=dict(rangebreaks=kwargs.get("rangebreaks", guess_plotly_rangebreaks(group_mean.index))),
            hovermode='x unified',
            hoverlabel=dict(bgcolor='rgba(255, 255, 255, 0.5)'),
        ),
    ).figure

    long_demeaned = long_demeaned.select("date", "time", pl.col("label").alias("long-average")).group_by("date",).mean().sort(by="date").drop("time")
    t_df = t_df.pivot(on="quantile", index="date", values="label").join(long_demeaned, on="date", how="left")
    t_df = t_df.with_columns(
        (pl.col(str(N)) - pl.col("1")).alias("long-short"),
    ).to_pandas().set_index("date", drop=True)
    t_df.rename(columns={i: f"Group{i}" for i in range(1, N+1)}, inplace=True)
    t_df.index = pd.to_datetime(t_df.index)

    t_df = t_df.dropna(how="all")  # for days which does not contain label
    # Cumulative Return By Group
    group_scatter_figure = ScatterGraph(
        t_df.cumsum(),
        layout=dict(
            title="Cumulative Return",
            xaxis=dict(tickangle=45, rangebreaks=kwargs.get("rangebreaks", guess_plotly_rangebreaks(t_df.index))),
            hovermode='x unified',
            hoverlabel=dict(bgcolor='rgba(255, 255, 255, 0.5)'),
        ),
    ).figure

    t_df = t_df.loc[:, ["long-short", "long-average"]]
    _bin_size = float(((t_df.max() - t_df.min()) / 20).min())
    group_hist_figure = SubplotsGraph(
        t_df,
        kind_map=dict(kind="DistplotGraph", kwargs=dict(bin_size=_bin_size)),
        subplots_kwargs=dict(
            rows=1,
            cols=2,
            print_grid=False,
            subplot_titles=["long-short", "long-average"],
        ),
    ).figure

    return group_bar_figure, group_scatter_figure, group_hist_figure


def _plot_qq(data: pd.Series = None, dist=stats.norm) -> go.Figure:
    """

    :param data:
    :param dist:
    :return:
    """
    # NOTE: plotly.tools.mpl_to_plotly not actively maintained, resulting in errors in the new version of matplotlib,
    # ref: https://github.com/plotly/plotly.py/issues/2913#issuecomment-730071567
    # removing plotly.tools.mpl_to_plotly for greater compatibility with matplotlib versions
    _plt_fig = sm.qqplot(data.dropna(), dist=dist, fit=True, line="45")
    plt.close(_plt_fig)
    qqplot_data = _plt_fig.gca().lines
    fig = go.Figure()

    fig.add_trace(
        {
            "type": "scatter",
            "x": qqplot_data[0].get_xdata(),
            "y": qqplot_data[0].get_ydata(),
            "mode": "markers",
            "marker": {"color": "#19d3f3"},
        }
    )

    fig.add_trace(
        {
            "type": "scatter",
            "x": qqplot_data[1].get_xdata(),
            "y": qqplot_data[1].get_ydata(),
            "mode": "lines",
            "line": {"color": "#636efa"},
        }
    )
    del qqplot_data
    fig.update_layout(**dict(hovermode='x unified', hoverlabel=dict(bgcolor='rgba(255, 255, 255, 0.5)'),))
    return fig


def _pred_ic(
        pred_label: pl.DataFrame = None, methods: Sequence[Literal["IC", "Rank IC"]] = ("IC", "Rank IC"), **kwargs
) -> tuple:
    """

    :param pred_label: pd.DataFrame
    must contain one column of realized return with name `label` and one column of predicted score names `score`.
    :param methods: Sequence[Literal["IC", "Rank IC"]]
    IC series to plot.
    IC is sectional pearson correlation between label and score
    Rank IC is the spearman correlation between label and score
    For the Monthly IC, IC histogram, IC Q-Q plot.  Only the first type of IC will be plotted.
    :return:
    """
    _methods_mapping = {"IC": "pearson", "Rank IC": "spearman"}
    grouper = ["date", "time"]

    ic_df = pred_label.drop_nulls(subset=grouper).group_by(grouper).agg(
        pl.corr("score", "label", method=_methods_mapping[m]).alias(m) for m in methods
    ).group_by("date").mean().drop("time").sort(by="date")
    ic_df = ic_df.to_pandas().set_index("date", drop=True)
    ic_df.index = pd.to_datetime(ic_df.index)

    _ic = ic_df.iloc(axis=1)[0]

    _index = _ic.index.get_level_values(0).astype("str").str.replace("-", "").str.slice(0, 6)
    _monthly_ic = _ic.groupby(_index).mean()
    _monthly_ic.index = pd.MultiIndex.from_arrays(
        [_monthly_ic.index.str.slice(0, 4), _monthly_ic.index.str.slice(4, 6)],
        names=["year", "month"],
    )

    # fill month
    _month_list = pd.date_range(
        start=pd.Timestamp(f"{_index.min()[:4]}0101"),
        end=pd.Timestamp(f"{_index.max()[:4]}1231"),
        freq="1ME",
    )
    _years = []
    _month = []
    for _date in _month_list:
        _date = _date.strftime("%Y%m%d")
        _years.append(_date[:4])
        _month.append(_date[4:6])

    fill_index = pd.MultiIndex.from_arrays([_years, _month], names=["year", "month"])

    _monthly_ic = _monthly_ic.reindex(fill_index)

    ic_bar_figure = ic_figure(ic_df, kwargs.get("show_nature_day", False))

    ic_heatmap_figure = HeatmapGraph(
        _monthly_ic.unstack(),
        layout=dict(title="Monthly IC",
                    xaxis=dict(dtick=1),
                    yaxis=dict(tickformat="04d", dtick=1),
                    hovermode='x unified',
                    hoverlabel=dict(bgcolor='rgba(255, 255, 255, 0.5)'),),
        graph_kwargs=dict(xtype="array", ytype="array"),
    ).figure

    dist = stats.norm
    _qqplot_fig = _plot_qq(_ic, dist)

    if isinstance(dist, stats.norm.__class__):
        dist_name = "Normal"
    else:
        dist_name = "Unknown"

    _ic_df = _ic.to_frame("IC")
    _bin_size = ((_ic_df.max() - _ic_df.min()) / 20).min()
    _sub_graph_data = [
        (
            "IC",
            dict(
                row=1,
                col=1,
                name="",
                kind="DistplotGraph",
                graph_kwargs=dict(bin_size=_bin_size),
            ),
        ),
        (_qqplot_fig, dict(row=1, col=2)),
    ]
    ic_hist_figure = SubplotsGraph(
        _ic_df.dropna(),
        kind_map=dict(kind="HistogramGraph", kwargs=dict()),
        subplots_kwargs=dict(
            rows=1,
            cols=2,
            print_grid=False,
            subplot_titles=["IC", "IC %s Dist. Q-Q" % dist_name],
        ),
        sub_graph_data=_sub_graph_data,
        layout=dict(
            yaxis2=dict(title="Observed Quantile"),
            xaxis2=dict(title=f"{dist_name} Distribution Quantile"),
        ),
    ).figure

    return ic_bar_figure, ic_heatmap_figure, ic_hist_figure

def _pred_autocorr(pred_label: pl.DataFrame, lag=1, **kwargs) -> tuple:
    dateList = pred_label["date"].unique(maintain_order=True)
    timeList = pred_label["time"].unique(maintain_order=True)
    asset_length = pred_label["asset"].unique().len()
    reshape = (-1, asset_length)
    arr = pred_label["score"].to_numpy().reshape(reshape)
    shifted = np.full_like(arr, np.nan)

    time_length = timeList.len()
    lag = lag * time_length

    if lag > 0:
        shifted[lag:, :] = arr[:-lag, :]  # 向下移动
    else:
        shifted[:lag, :] = arr[-lag:, :]  # 向上移动
    corr = pd.DataFrame(arr).corrwith(pd.DataFrame(shifted), method="spearman", axis=1)
    index = dateList.to_frame().join(timeList.to_frame(), how="cross")
    _df = index.with_columns(pl.Series(name="value", values=corr.values,)).group_by("date").mean().drop("time").sort(by="date")
    _df = _df.to_pandas().set_index("date", drop=True)
    _df.index = pd.to_datetime(_df.index)
    ac_figure = ScatterGraph(
        _df,
        layout=dict(
            title="Auto Correlation",
            xaxis=dict(tickangle=45, rangebreaks=kwargs.get("rangebreaks", guess_plotly_rangebreaks(_df.index))),
            hovermode='x unified',
            hoverlabel=dict(bgcolor='rgba(255, 255, 255, 0.5)'),
        ),
    ).figure
    return (ac_figure,)

def _pred_turnover(pred_label: pl.DataFrame, N=5, lag=1, **kwargs) -> tuple:
    pred_label = pred_label.with_columns(
        pl.col("score").qcut(N,
                             labels=[str(i) for i in range(1, N+1)],
                             allow_duplicates=True).over(["date", "time"]).alias("quantile")
    ).with_columns(
        pl.col(pl.Categorical).cast(pl.Int32))

    pred_label = pred_label.drop_nulls(subset="quantile").filter(((pl.col("quantile")==1) | (pl.col("quantile") == N)))
    grouper = ["quantile", "date", "time"]
    cur_asset_set = pred_label.group_by(grouper).agg(
        pl.col("asset").unique()
    ).sort(by=grouper)
    shift_grouper = ["quantile", "time"]
    prev_asset_set = cur_asset_set.select(
        *grouper,
        pl.col("asset").shift(lag).over(shift_grouper)
    )
    r_df = cur_asset_set.join(
        prev_asset_set, on=grouper, how="left", suffix="_prev"
    ).with_columns(
        pl.col("asset").list.set_difference("asset_prev").alias("asset_diff").list.len().alias("diff_count"),
        pl.col("asset").list.len().alias("cur_count"),
    ).select(
        # "factor_quantile",
        "quantile", "date", "time",
        (pl.col("diff_count") / pl.col("cur_count")).alias("turnover")
    )
    r_df = r_df.group_by("quantile", "date").mean().drop("time").sort(by="date")
    r_df = r_df.pivot(on="quantile", index="date", values="turnover").to_pandas().set_index("date", drop=True)
    r_df.rename(columns={"1": "Group1", str(N): f"Group{N}"}, inplace=True)
    r_df.index = pd.to_datetime(r_df.index)

    turnover_figure = ScatterGraph(
        r_df,
        layout=dict(
            title="Top-Bottom Turnover",
            xaxis=dict(tickangle=45, rangebreaks=kwargs.get("rangebreaks", guess_plotly_rangebreaks(r_df.index))),
            hovermode='x unified',
            hoverlabel=dict(bgcolor='rgba(255, 255, 255, 0.5)'),
        ),
    ).figure
    return (turnover_figure,)


def ic_figure(ic_df: pd.DataFrame, show_nature_day=True, **kwargs) -> go.Figure:
    r"""IC figure

    :param ic_df: ic DataFrame
    :param show_nature_day: whether to display the abscissa of non-trading day
    :param \*\*kwargs: contains some parameters to control plot style in plotly. Currently, supports
       - `rangebreaks`: https://plotly.com/python/time-series/#Hiding-Weekends-and-Holidays
    :return: plotly.graph_objs.Figure
    """
    if show_nature_day:
        date_index = pd.date_range(ic_df.index.min(), ic_df.index.max())
        ic_df = ic_df.reindex(date_index)
    ic_bar_figure = BarGraph(
        ic_df,
        layout=dict(
            title="Information Coefficient (IC)",
            xaxis=dict(tickangle=45, rangebreaks=kwargs.get("rangebreaks", guess_plotly_rangebreaks(ic_df.index))),
            hovermode='x unified',
            hoverlabel=dict(bgcolor='rgba(255, 255, 255, 0.5)'),
        ),
    ).figure
    return ic_bar_figure


def model_performance_graph(
    factor_data: pl.DataFrame,
    lag: int = 1,
    N: int = 5,
    graph_names: list = ["group_return", "pred_ic",],
    show_notebook: bool = True,
    show_nature_day: bool = False,
    **kwargs,
) -> [list, tuple]:
    r"""Model performance

    :param pred_label: index is **pd.MultiIndex**, index name is **[instrument, datetime]**; columns names is **[score, label]**.
           It is usually same as the label of model training(e.g. "Ref($close, -2)/Ref($close, -1) - 1").


            .. code-block:: python

                instrument  datetime        score       label
                SH600004    2017-12-11  -0.013502       -0.013502
                                2017-12-12  -0.072367       -0.072367
                                2017-12-13  -0.068605       -0.068605
                                2017-12-14  0.012440        0.012440
                                2017-12-15  -0.102778       -0.102778


    :param lag: `pred.groupby(level='instrument')['score'].shift(lag)`. It will be only used in the auto-correlation computing.
    :param N: group number, default 5.
    :param graph_names: graph names; default ['cumulative_return', 'pred_ic', 'pred_autocorr', 'pred_turnover'].
    :param show_notebook: whether to display graphics in notebook, the default is `True`.
    :param show_nature_day: whether to display the abscissa of non-trading day.
    :param \*\*kwargs: contains some parameters to control plot style in plotly. Currently, supports
       - `rangebreaks`: https://plotly.com/python/time-series/#Hiding-Weekends-and-Holidays
    :return: if show_notebook is True, display in notebook; else return `plotly.graph_objs.Figure` list.
    """
    ylog.info("多周期共用指标")
    common_figs = list()
    # label已经在 factor_data 层面处理过，所以 long_short/group_neutral 均设置为 False
    summary_tb = tears.get_summary_report(factor_data,
                                          long_short=False,
                                          group_neutral=False,
                                          by_time=True)
    table = finplot.table(summary_tb.to_pandas(), highlight_cols=["ic", "top_bps", "bottom_bps", "spread_bps"])
    common_figs.append(table)
    pred_label = factor_data.select("date", "time", "asset", pl.col("factor").alias("score"),)
    for graph_name in ["pred_autocorr", "pred_turnover"]:
        fun_res = eval(f"_{graph_name}")(
            pred_label=pred_label, lag=lag, N=N, show_nature_day=show_nature_day, **kwargs
        )
        common_figs += fun_res
    if show_notebook:
        BarGraph.show_graph_in_notebook(common_figs)
    periods = utils.get_forward_returns_columns(factor_data.columns)
    figure_dict = dict()

    for period in periods:
        ylog.info(f"{period} 分析")
        pred_label = factor_data.select("date", "time", "asset", pl.col("factor").alias("score"), pl.col(period).alias("label"), pl.col("factor_quantile").alias("quantile"))
        figure_list = []
        for graph_name in graph_names:
            fun_res = eval(f"_{graph_name}")(
                pred_label=pred_label, lag=lag, N=N, show_nature_day=show_nature_day, **kwargs
            )
            figure_list += fun_res

        if show_notebook:
            BarGraph.show_graph_in_notebook(figure_list)
        figure_dict[period] = figure_list
    return common_figs, figure_dict



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Created on 2025/3/5 21:40
@author: ZhangYundi
@email: yundi.xxii@outlook.com
---------------------------------------------
"""
from __future__ import annotations

import math
from functools import wraps

from .qdf import QDF
from .lazy import LQDF


def signature(function):
    @wraps(function)
    def wrapper(*args, **kwargs):
        args = [str(arg) for arg in args]
        for k, v in kwargs.items():
            args.append(f"{k}={v}")
        string = f"{function.__name__}({', '.join(args)})"
        return string

    return wrapper


def from_polars(df, index: tuple[str] = ("date", "time", "asset"), align: bool = True, ) -> QDF:
    """polars dataframe 转为 表达式数据库"""
    return QDF(df, index, align,)

def to_lazy(df, index: tuple[str] = ("date", "time", "asset"), align: bool = False, ) -> LQDF:
    """polars dataframe 转为 表达式数据库"""
    return LQDF(df, index, align,)


@signature
def abs(expr: str): ...


@signature
def log(expr: str, base=math.e): ...


@signature
def sqrt(expr: str): ...


@signature
def square(expr: str): ...


@signature
def cube(expr: str): ...


@signature
def cbrt(expr: str): ...


@signature
def sin(expr: str): ...


@signature
def sinh(expr: str): ...


@signature
def arcsin(expr: str): ...


@signature
def arcsinh(expr: str): ...


@signature
def cos(expr: str): ...


@signature
def cosh(expr: str): ...


@signature
def arccos(expr: str): ...


@signature
def arccosh(expr: str): ...


@signature
def tan(expr: str): ...


@signature
def tanh(expr: str): ...


@signature
def arctan(expr: str): ...


@signature
def arctanh(expr: str): ...


@signature
def sign(expr: str): ...


@signature
def sigmoid(expr: str): ...


@signature
def cot(expr: str): ...


@signature
def degrees(expr: str): ...


@signature
def ind_entropy(expr: str, windows: int): ...

@signature
def d_entropy(expr: str, windows: int): ...

@signature
def cs_entropy(expr: str, windows: int): ...

@signature
def ts_entropy(expr: str, windows: int): ...


@signature
def exp(expr: str): ...


@signature
def log1p(expr: str): ...


@signature
def clip(expr: str, lower_bound, upper_bound): ...


@signature
def cs_ufit(expr: str): ...


@signature
def cs_rank(expr: str): ...


@signature
def cs_demean(expr: str): ...


@signature
def cs_mean(expr: str): ...


@signature
def cs_mid(expr: str): ...


@signature
def cs_moderate(expr: str): ...


@signature
def cs_qcut(expr: str, N: int = 10): ...


@signature
def cs_ic(left: str, right: str): ...


@signature
def cs_corr(left: str, right: str): ...


@signature
def cs_std(expr: str): ...


@signature
def cs_var(expr: str): ...


@signature
def cs_skew(expr: str): ...


@signature
def cs_slope(left: str, right: str): ...


@signature
def cs_resid(left: str, right: str): ...


@signature
def cs_zscore(expr: str): ...


@signature
def cs_midby(expr: str, *by: str): ...


@signature
def cs_meanby(expr: str, *by: str): ...


@signature
def cs_max(expr: str): ...


@signature
def cs_min(expr: str): ...


@signature
def cs_peakmax(expr: str): ...


@signature
def cs_peakmin(expr: str): ...


@signature
def ts_mean(expr: str, windows: int): ...


@signature
def ts_std(expr: str, windows: int): ...


@signature
def ts_sum(expr: str, windows: int): ...


@signature
def ts_var(expr: str, windows: int): ...


@signature
def ts_skew(expr: str, windows: int): ...


@signature
def ts_ref(expr: str, windows: int): ...


@signature
def ts_mid(expr: str, windows: int): ...


@signature
def ts_mad(expr: str, windows: int): ...


@signature
def ts_rank(expr: str, windows: int): ...


@signature
def ts_max(expr: str, windows: int): ...


@signature
def ts_min(expr: str, windows: int): ...


@signature
def ts_ewmmean(expr: str, com=None, span=None, half_life=None, alpha=None): ...


@signature
def ts_ewmstd(expr: str, com=None, span=None, half_life=None, alpha=None): ...


@signature
def ts_ewmvar(expr: str, com=None, span=None, half_life=None, alpha=None): ...


@signature
def ts_cv(expr: str, windows: int): ...


@signature
def ts_snr(expr: str, windows: int): ...


@signature
def ts_diff(expr: str, windows: int): ...


@signature
def ts_pct(expr: str, windows: int): ...


@signature
def ts_slope(left: str, right: str, windows: int): ...


@signature
def ts_corr(left: str, right: str, windows: int): ...


@signature
def ts_cov(left: str, right: str, windows: int): ...


@signature
def ts_resid(left: str, right: str, windows: int): ...


@signature
def ts_quantile(expr: str, windows: int, q: float): ...


@signature
def ts_prod(expr: str, windows: int): ...


@signature
def d_mean(expr: str, windows: int): ...


@signature
def d_std(expr: str, windows: int): ...


@signature
def d_sum(expr: str, windows: int): ...


@signature
def d_var(expr: str, windows: int): ...


@signature
def d_skew(expr: str, windows: int): ...


@signature
def d_ref(expr: str, windows: int): ...


@signature
def d_mid(expr: str, windows: int): ...


@signature
def d_mad(expr: str, windows: int): ...


@signature
def d_rank(expr: str, windows: int): ...


@signature
def d_max(expr: str, windows: int): ...


@signature
def d_min(expr: str, windows: int): ...


@signature
def d_ewmmean(expr: str, com=None, span=None, half_life=None, alpha=None): ...


@signature
def d_ewmstd(expr: str, com=None, span=None, half_life=None, alpha=None): ...


@signature
def d_ewmvar(expr: str, com=None, span=None, half_life=None, alpha=None): ...


@signature
def d_cv(expr: str, windows: int): ...


@signature
def d_snr(expr: str, windows: int): ...


@signature
def d_diff(expr: str, windows: int): ...


@signature
def d_pct(expr: str, windows: int): ...


@signature
def d_slope(left: str, right: str, windows: int): ...


@signature
def d_corr(left: str, right: str, windows: int): ...


@signature
def d_cov(left: str, right: str, windows: int): ...


@signature
def d_resid(left: str, right: str, windows: int): ...


@signature
def d_quantile(expr: str, windows: int, q: float): ...


@signature
def d_prod(expr: str, windows: int): ...


@signature
def ind_mean(expr: str, windows: int): ...


@signature
def ind_std(expr: str, windows: int): ...


@signature
def ind_sum(expr: str, windows: int): ...


@signature
def ind_var(expr: str, windows: int): ...


@signature
def ind_skew(expr: str, windows: int): ...


@signature
def ind_ref(expr: str, windows: int): ...


@signature
def ind_mid(expr: str, windows: int): ...


@signature
def ind_mad(expr: str, windows: int): ...


@signature
def ind_rank(expr: str, windows: int): ...


@signature
def ind_max(expr: str, windows: int): ...


@signature
def ind_min(expr: str, windows: int): ...


@signature
def ind_ewmmean(expr: str, com=None, span=None, half_life=None, alpha=None): ...


@signature
def ind_ewmstd(expr: str, com=None, span=None, half_life=None, alpha=None): ...


@signature
def ind_ewmvar(expr: str, com=None, span=None, half_life=None, alpha=None): ...


@signature
def ind_cv(expr: str, windows: int): ...


@signature
def ind_snr(expr: str, windows: int): ...


@signature
def ind_diff(expr: str, windows: int): ...


@signature
def ind_pct(expr: str, windows: int): ...


@signature
def ind_slope(left: str, right: str, windows: int): ...


@signature
def ind_corr(left: str, right: str, windows: int): ...


@signature
def ind_cov(left: str, right: str, windows: int): ...


@signature
def ind_resid(left: str, right: str, windows: int): ...


@signature
def ind_quantile(expr: str, windows: int, q: float): ...


@signature
def ind_prod(expr: str, windows: int): ...


@signature
def max(*exprs: str): ...


@signature
def min(*exprs: str): ...


@signature
def sum(*exprs: str): ...



# -*- coding: utf-8 -*-
"""
---------------------------------------------
Copyright (c) 2025 ZhangYundi
Licensed under the MIT License. 
Created on 2025/6/26 15:54
Email: yundi.xxii@outlook.com
Description: quda cli 入口
---------------------------------------------
"""

import typer

app = typer.Typer()

@app.callback(invoke_without_command=True)
def main(ctx: typer.Context):
    if ctx.invoked_subcommand is None:
        typer.echo("Run `quda --help` to get more information.")

@app.command()
def init_config():

    """生成模版配置，复制自: quda/ml/conf"""

    typer.echo(f"[QUDA] - Copy quda.ml.conf template.")

    from .app import init_config as init_config_
    init_config_()

@app.command()
def update():

    """数据更新"""
    import quda

    quda.update()

@app.command()
def tick_clean(beg_date: str, end_date: str, tb_name: str = "mc/stock_ytick", n_jobs: int = 5):
    """tick 行情数据清洗"""

    from quda.data.quote import save_ytick

    save_ytick(tb_name=tb_name, beg_date=beg_date, end_date=end_date, n_jobs=n_jobs)




if __name__ == '__main__':
    app()